"""Contains high-level object representing the whole of an IOS Shell file."""
from dataclasses import dataclass
import datetime
from typing import Any, Dict, List, Optional, Union

from . import parsing, sections, utils


@dataclass
class ShellFile:
    """Represents the contents of an IOS Shell file."""

    filename: str
    """The file name the data originally came from.
    Useful for debugging."""
    modified_date: datetime.datetime
    header_version: sections.Version
    file: sections.FileInfo
    administration: sections.Administration
    location: sections.Location
    instrument: Optional[sections.Instrument]
    history: Optional[sections.History]
    calibration: Optional[sections.Calibration]
    deployment: Optional[sections.Deployment]
    recovery: Optional[sections.Recovery]
    raw: Optional[sections.Raw]
    """The contents of the section starting with \\*RAW, not the raw file contents."""
    comments: str
    data: Union[List[List[Any]], str]
    """Data is assumed to be processed if it is a List[List[Any]],
    and unprocessed if it is a str.

    Some data has been known to contain dates or arbitrary strings, so raw numpy arrays
    are difficult to make work."""
    geo_code: str = "None"

    @classmethod
    def fromfile(cls, filename, process_data=True):  # pragma: no mutate
        """Construct a ShellFile object from the contents of a file."""
        with open(filename, "r", encoding="ASCII", errors="ignore") as f:
            contents = f.read()
        try:
            return ShellFile.fromcontents(contents, process_data, filename=filename)
        except ValueError as e:
            if filename in str(e):  # pragma: no mutate
                raise
            exc = ValueError(f"Error in {filename}: {e}")  # pragma: no mutate
            exc.__traceback__ = e.__traceback__  # pragma: no mutate
            raise exc from None

    @classmethod
    def fromcontents(cls, contents, process_data=True, filename="bare string"):
        """Construct a ShellFile object from the contents of a string."""
        header, raw_data = contents.split("*END OF HEADER", 1)  # pragma: no mutate
        header_lines = header.splitlines()
        header_lines.append("*END OF HEADER")
        while "" in header_lines:
            header_lines.remove("")
        modified_date, rest = parsing.get_modified_date(header_lines)
        header_version, rest = parsing.get_header_version(rest)
        # begin named sections
        # sections may appear out of order
        (
            file,
            administration,
            location,
            instrument,
            history,
            calibration,
            deployment,
            recovery,
            comments,
            raw,
        ) = (
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )
        line = rest[0]
        while not line.startswith("*END OF HEADER"):
            if line.startswith("*FILE"):
                if file is not None:
                    raise ValueError("There should only be one file section")
                file, rest = parsing.get_file(rest)
            elif line.startswith("*ADMINISTRATION"):
                if administration is not None:
                    raise ValueError("There should only be one administration section")
                administration, rest = parsing.get_administration(rest)
            elif line.startswith("*LOCATION"):
                if location is not None:
                    raise ValueError("There should only be one location section")
                location, rest = parsing.get_location(rest)
            elif line.startswith("*INSTRUMENT"):
                if instrument is not None:
                    raise ValueError("There should only be one instrument section")
                instrument, rest = parsing.get_instrument(rest)
            elif line.startswith("*HISTORY"):
                if history is not None:
                    raise ValueError("There should only be one history section")
                history, rest = parsing.get_history(rest)
            elif line.startswith("*COMMENTS"):
                if comments is not None:
                    raise ValueError("There should only be one comments section")
                comments, rest = parsing.get_comments(rest)
            elif line.startswith("*CALIBRATION"):
                if calibration is not None:
                    raise ValueError("There should only be one calibration section")
                calibration, rest = parsing.get_calibration(rest)
            elif line.startswith("*RAW"):
                if raw is not None:
                    raise ValueError("There should only be one raw section")
                raw, rest = parsing.get_raw(rest)
            elif line.startswith("*DEPLOYMENT"):
                if deployment is not None:
                    raise ValueError("There should only be one deployment section")
                deployment, rest = parsing.get_deployment(rest)
            elif line.startswith("*RECOVERY"):
                if recovery is not None:
                    raise ValueError("There should only be one recovery section")
                recovery, rest = parsing.get_recovery(rest)
            else:
                raise ValueError(f"Unknown section: {line}")
            line = rest[0]
        # end named sections
        # check for required sections
        if file is None:
            raise ValueError("*FILE section must be present")
        if administration is None:
            raise ValueError("*ADMINISTRATION section must be present")
        if location is None:
            raise ValueError("*LOCATION section must be present")
        info = ShellFile(
            filename=filename,
            modified_date=modified_date,
            header_version=header_version,
            file=file,
            administration=administration,
            location=location,
            instrument=instrument,
            history=history,
            calibration=calibration,
            deployment=deployment,
            recovery=recovery,
            raw=raw,
            comments=comments if comments is not None else "",
            data=raw_data,
        )
        if process_data:
            info.process_data()
        return info

    def get_location(self) -> Dict[str, float]:
        """Produce a conventent dict of the longitude and latitude in the file."""
        return {
            "longitude": self.location.longitude,
            "latitude": self.location.latitude,
        }

    def get_time(self) -> datetime.datetime:
        """A time to associate the contents with.

        Typically the START TIME field of \\*FILE
        """
        if self.file.start_time != datetime.datetime.min:
            return self.file.start_time
        elif self.file.end_time != datetime.datetime.min:
            return self.file.end_time
        else:
            raise ValueError("No valid time found")

    def data_is_processed(self) -> bool:
        """Determine whether or not the data in the file has already been processed."""
        return not isinstance(self.data, str)

    def process_data(self):
        """Perform the processing of the data."""
        if self.data_is_processed():
            return
        # assertion to satisfy (optional) type checking
        assert isinstance(self.data, str)
        try:
            data, _ = parsing.get_data(
                self.data, self.file.format, self.file.number_of_records
            )
            data = [self.__insert_pads(row) for row in data]
            self.data = data
        except Exception as e:
            exc = ValueError(
                f"Error processing data in {self.filename}: {e}"  # pragma: no mutate
            )  # pragma: no mutate
            exc.__traceback__ = e.__traceback__  # pragma: no mutate
            raise exc from None

    def get_complete_header(self) -> Dict[str, Union[Dict[str, Any], str]]:
        """Returns all the present header information contained in the source file"""
        header: Dict[str, Union[Dict[str, Any], str]] = {}
        if self.file is not None:
            header["file"] = self.file.raw
        if self.administration is not None:
            header["administration"] = self.administration.raw
        if self.location is not None:
            header["location"] = self.location.raw
        if self.instrument is not None:
            header["instrument"] = self.instrument.raw
        if self.history is not None:
            header["history"] = self.history.raw
        if self.calibration is not None:
            header["calibration"] = self.calibration.raw
        if self.deployment is not None:
            header["deployment"] = self.deployment.raw
        if self.recovery is not None:
            header["recovery"] = self.recovery.raw
        if self.raw is not None:
            header["raw"] = self.raw.raw
        if self.comments is not None:
            header["comments"] = self.comments
        return header

    def get_obs_time(self) -> List[datetime.datetime]:
        """Returns a list of all the datetime objects associated with the data"""
        utc = datetime.timezone.utc

        channel_names = [channel.name.strip().lower() for channel in self.file.channels]

        if "date" in channel_names and "time" in channel_names:
            date_idx = channel_names.index("date")
            time_idx = channel_names.index("time")
            dates = [row[date_idx] for row in self.data]
            times = [row[time_idx] for row in self.data]
            datetimes = [
                datetime.datetime.combine(date, time)
                for date, time in zip(dates, times)
            ]
            obs_time = [i.replace(tzinfo=utc) for i in datetimes]
        elif "date_time" in channel_names:
            date_idx = channel_names.index("date_time")
            dates = [row[date_idx] for row in self.data]
            obs_time = [i.replace(tzinfo=utc) for i in dates]
        else:
            # no date/time columns, assume TIME INCREMENT is present
            try:
                time_increment = self.file.time_increment
                obs_time = [
                    self.file.start_time + (time_increment * i)
                    for i in range(self.file.number_of_records)
                ]
                if utils.has_many_values(obs_time) and utils.all_same(obs_time):
                    raise Exception()
            except Exception as e:
                raise Exception("ERROR: Unable to use time increment", self.filename)

        # date/time section in data is supposed to be in UTC.
        # check if they match, if not then raise fatal error

        if obs_time[0] != self.file.start_time:
            raise Exception(
                f"Error: First record in data ({obs_time[0]}) does not match start date in header ({self.file.start_time})",  # pragma: no mutate
                self.filename,
            )
        return obs_time

    def to_pandas(self):
        """Convert processed data to pandas.DataFrame

        :return: a pandas.DataFrame if pandas is present and data has been processed, otherwise None
        """
        if not self.data_is_processed():
            return None

        names = [channel.name for channel in self.file.channels]
        return utils.list_to_pandas(self.data, names)

    def __insert_pads(self, row):
        return [
            item if str(item) != "" else self.file.channel_details[i].pad
            for i, item in enumerate(row)
        ]



"""This module holds constants for expected keys for each section in an IOS Shell file.

It exists because it is easier to ensure consistent string values by using constants instead of raw strings.
This also means typos are easier to correct.
"""

# common to all sections
REMARKS = "remarks"

# *FILE keys
START_TIME = "start time"
END_TIME = "end time"
TIME_ZERO = "time zero"
TIME_INCREMENT = "time increment"
TIME_UNITS = "time units"
NUMBER_OF_RECORDS = "number of records"
DATA_DESCRIPTION = "data description"
FILE_TYPE = "file type"
DATA_TYPE = "data type"
NUMBER_OF_CHANNELS = "number of channels"
CHANNELS = "channels"
CHANNEL_DETAIL = "channel detail"
FORMAT = "format"
CONTINUED = "continued"
PAD = "pad"

# *ADMINISTRATION keys
MISSION = "mission"
AGENCY = "agency"
COUNTRY = "country"
PROJECT = "project"
SCIENTIST = "scientist"
PLATFORM = "platform"

# *LOCATION keys
GEOGRAPHIC_AREA = "geographic area"
STATION = "station"
EVENT_NUMBER = "event number"
LATITUDE = "latitude"
LONGITUDE = "longitude"
WATER_DEPTH = "water depth"

# *INSTRUMENT keys
TYPE = "type"
MODEL = "model"
SERIAL_NUMBER = "serial number"
DEPTH = "depth"

# *HISTORY keys
PROGRAMS = "programs"

# *CALIBRATION keys
CORRECTED_CHANNELS = "corrected channels"

# *DEPLOYMENT keys
TIME_ANCHOR_DROPPED = "time anchor dropped"

# *RECOVERY keys
TIME_ANCHOR_RELEASED = "time anchor released"



"""Contains useful functions for parsing that are not themselves parsing functions."""
import datetime
from typing import List, Any

from .regex import *


def apply_column_mask(data: str, mask: str) -> List[str]:
    """Apply a mask to a single row of data.

    :param data: the row of data to break up
    :param mask: a string with '-' for every character to be included as an element
    """
    data = data.ljust(len(mask))
    end = 0
    out = []
    while end > -1:  # pragma: no mutate
        start = mask.find("-", end)  # pragma: no mutate
        end = mask.find(" ", start)
        out.append(data[start:end])
    out[-1] += data[end:]
    return out


def format_string(format: str, kind: str, width: int, decimals: int) -> str:
    """Construct an appropriate format string for the given type.

    :param format: the format the data is expected to conform to
    :param kind: the type the data is expected to be
    :param width: the number of characters the data may take up
    :param decimals: the number of characters after a decimal a float is intended to use
    """
    fortrantype = format.strip().upper()
    datatype = kind.strip().upper()
    if datatype in ["DT"]:
        return f"A17"
    elif fortrantype in ["F"]:
        return f"F{width}.{decimals}"
    elif fortrantype in ["E"]:
        return f"E{width}.{decimals}"
    elif fortrantype in ["I"]:
        return f"I{width}"
    elif fortrantype.upper() in [
        "YYYY/MM/DD",
        "YYYY-MM-DD",
        "HH:MM",
        "HH:MM:SS",
        "HH:MM:SS.SS",
    ]:
        return f"A{len(fortrantype)+1}"
    elif fortrantype in ["' '", "NQ"]:
        return f"A{width}"
    else:
        return fortrantype


def _to_timezone_offset(name: str) -> int:
    upper = name.upper()
    if upper in ["UTC", "GMT"]:
        return 0
    elif upper in ["ADT"]:
        return -3
    elif upper in ["MDT", "CST"]:
        return -6
    elif upper in ["PDT", "MST"]:
        return -7
    elif upper in ["PST"]:
        return -8
    else:
        raise ValueError(f"Unknown time zone: {name}.")


def to_date(contents: str) -> datetime.date:
    """Convert a string representing a date into a usable object."""
    date_info = [int(part) for part in contents.strip().replace("-", "/").split("/")]
    year = date_info[0]
    month = date_info[1]
    day = date_info[2]
    return datetime.date(year, month, day)


def to_time(contents: str, tzinfo=datetime.timezone.utc) -> datetime.time:
    """Convert a string representing a time into a usable object."""
    time_info = [
        int(part) for piece in contents.strip().split(":") for part in piece.split(".")
    ]
    hour = time_info[0]
    minute = time_info[1]
    second = time_info[2] if len(time_info) > 2 else 0
    return datetime.time(hour=hour, minute=minute, second=second, tzinfo=tzinfo)


def _to_datetime(tz: str, date: str, time: str) -> datetime.datetime:
    tzoffset = _to_timezone_offset(tz)
    date_obj = to_date(date)
    tz_obj = datetime.timezone(datetime.timedelta(hours=tzoffset))
    if time != "":
        time_obj = to_time(time, tz_obj)
        return datetime.datetime.combine(date_obj, time_obj)
    else:
        return datetime.datetime(
            date_obj.year, date_obj.month, date_obj.day, tzinfo=tz_obj
        )


def to_datetime(value: str) -> datetime.datetime:
    """Convert a string representing a date and time into a usable object."""
    # attempting to cover "Unknown" and "Unk.000"
    if value in ["", "?"] or "unk" in value.lower():
        return datetime.datetime.min
    trimmed = value.strip()
    # separate matches are required in order to avoid reusing group names
    if m := DATE_TIME_PATTERN.match(trimmed):
        return _to_datetime(tz="UTC", **m.groupdict())
    elif m := TZ_DATE_TIME_PATTERN.match(trimmed):
        return _to_datetime(**m.groupdict(""))
    elif m := DATE_TIME_TZ_PATTERN.match(trimmed):
        return _to_datetime(**m.groupdict(""))
    else:
        raise ValueError(f"Unknown time format: {value}")


def to_increment(description: str) -> datetime.timedelta:
    if description == "n/a":
        return datetime.timedelta(seconds=0)
    days, hours, minutes, seconds, mseconds = description.split(" ")
    return datetime.timedelta(
        days=float(days),
        hours=float(hours),
        minutes=float(minutes),
        seconds=float(seconds),
        milliseconds=float(mseconds),
    )


def _get_coord(raw_coord: str, positive_marker: str, negative_marker: str) -> float:
    coord = raw_coord.split("!")[0]
    degrees, minutes, direction = coord.split()
    out = float(degrees) + float(minutes) / 60.0
    if direction.upper() == positive_marker.upper():
        return out
    elif direction.upper() == negative_marker.upper():
        return out * -1.0  # pragma: no mutate
    else:
        raise ValueError("Coordinate contains unknown direction marker")


def get_latitude(coord: str) -> float:
    """Convert a string representing a latitude into a floating point number."""
    return _get_coord(coord, "N", "S")


def get_longitude(coord: str) -> float:
    """Convert a string representing a longitude into a floating point number."""
    return _get_coord(coord, "E", "W")


def is_section_heading(s: str) -> bool:
    """Decide whether or not a string represents the beginning of a secion."""
    return SECTION_HEADING_PATTERN.fullmatch(s) is not None


def is_table_mask(line: str) -> bool:
    """Decide whether or not a string is a table mask.

    A table mask starts with an indentation, then a '!' character, and then a mixture
    of ' ' and '-'
    """
    return (
        line.startswith(" ")
        and line.lstrip().startswith("!")
        and "-" in line
        and all(c in [" ", "!", "-"] for c in line)
    )


def has_many_values(list: List[Any]) -> bool:
    """Decide whether or not a list has more than one value"""
    return len(list) > 1


def all_same(list: List[Any]) -> bool:
    """Decide whether or not a list's values are all the same"""
    return len(set(list)) == 1


def list_to_pandas(list: List[List[Any]], names: List[str]):
    """Convert list of lists to pandas DataFrame

    :param list: the data as a list of data rows
    :param names: the row names to be used in the pandas.DataFrame
    :return: a pandas.DataFrame if pandas is present, otherwise None
    """
    try:
        import pandas
    except ImportError:
        return None  # pragma: no mutate

    df = pandas.DataFrame(list)
    return df.rename(columns=lambda old: names[old])



"""Read and process an IOS Shell file in Python."""

from .shell import ShellFile
from .sections import (
    Version,
    Channel,
    ChannelDetail,
    FileInfo,
    Administration,
    Location,
    Instrument,
    Program,
    Raw,
    History,
    Calibration,
    Deployment,
    Recovery,
)



"""Contains classes to represent complex elements of an IOS Shell file."""
from dataclasses import dataclass
import datetime
from typing import Any, Dict, List

EMPTY = ["", "' '", "n/a"]
NAN = float("NaN")


@dataclass
class Version:
    """Represents the \\*IOS HEADER VERSION line of the file."""

    version_no: str
    date1: str
    date2: str
    tag: str = ""


class Channel:
    """A single entry in the CHANNELS table."""

    no: int  # acts as an identifier
    name: str
    units: str
    minimum: float
    maximum: float

    def __init__(
        self,
        no="",
        name="",
        units="",
        minimum="",
        maximum="",
    ):
        self.no = int(no)
        self.name = name.strip()
        self.units = units.strip()
        # FIXME: remove when no longer relevant
        if minimum.strip().upper() == "O":
            minimum = "0"
        if maximum.strip().upper() == "O":
            maximum = "0"
        self.minimum = float(minimum) if minimum.strip() not in EMPTY else NAN
        self.maximum = float(maximum) if maximum.strip() not in EMPTY else NAN


class ChannelDetail:
    """A single entry in the CHANNEL DETAILS table."""

    no: int  # acts as an identifier
    pad: float
    start: str
    width: int
    format: str
    type: str
    decimal_places: int = -1

    def __init__(
        self,
        no="",
        pad="",
        start="",
        width="",
        format="",
        type="",
        decimal_places="",
    ):
        self.no = int(no)
        self.pad = float(pad) if pad.strip() not in EMPTY else NAN
        self.start = start
        self.width = int(width) if width.strip() not in EMPTY else 0
        self.format = format
        self.type = type
        self.decimal_places = (
            int(decimal_places) if decimal_places.strip() not in EMPTY else 0
        )


@dataclass
class FileInfo:
    """Represents the \\*FILE section."""

    start_time: datetime.datetime
    end_time: datetime.datetime
    time_zero: datetime.datetime
    time_increment: datetime.timedelta
    time_units: str
    number_of_records: int
    data_description: str
    file_type: str
    format: str
    data_type: str
    pad: float
    number_of_channels: int
    channels: List[Channel]
    channel_details: List[ChannelDetail]
    remarks: str
    raw: Dict[str, Any]


@dataclass
class Administration:
    """Represents the \\*ADMINISTRATION section."""

    mission: str
    agency: str
    country: str
    project: str
    scientist: str
    platform: str
    remarks: str
    raw: Dict[str, Any]


@dataclass
class Location:
    """Represents the \\*LOCATION section."""

    geographic_area: str
    station: str
    event_number: int
    latitude: float
    longitude: float
    water_depth: float
    remarks: str
    raw: Dict[str, Any]


@dataclass
class Instrument:
    """Represents the \\*INSTRUMENT section."""

    type: str
    model: str
    serial_number: str
    depth: float
    remarks: str
    raw: Dict[str, Any]


@dataclass
class Program:
    """A single entry in the PROGRAMS table."""

    name: str
    version: str
    date: datetime.date
    time: datetime.time
    records_in: int
    records_out: int


@dataclass
class Raw:
    """Represents the \\*RAW section."""

    remarks: str
    raw: Dict[str, Any]


@dataclass
class History:
    """Represents the \\*HISTORY section."""

    programs: List[Program]
    remarks: str
    raw: Dict[str, Any]


@dataclass
class Calibration:
    """Represents the \\*CALIBRATION section."""

    corrected_channels: List[Dict[str, str]]
    remarks: str
    raw: Dict[str, Any]


@dataclass
class Deployment:
    """Represents the \\*DEPLOYMENT section."""

    mission: str
    type: str
    anchor_dropped: datetime.datetime
    remarks: str
    raw: Dict[str, Any]


@dataclass
class Recovery:
    """Represents the \\*RECOVERY section."""

    mission: str
    anchor_released: datetime.datetime
    remarks: str
    raw: Dict[str, Any]



"""Contains functions for parsing files in IOS Shell format."""
import datetime
import fortranformat as ff
import itertools
import math
from typing import Any, Dict, List, Tuple

from . import sections, utils
from .regex import *
from .keys import *


def _next_line(rest: List[str]) -> Tuple[str, List[str]]:
    return rest[0], rest[1:]


def _has_key_prefix(line: str) -> bool:
    idx = line.find(":")
    key = line[:idx]
    return idx > 0 and KEY_PREFIX_PATTERN.fullmatch(key)


def get_modified_date(contents: List[str]) -> Tuple[datetime.datetime, List[str]]:
    """Parse the modified date."""
    line, rest = _next_line(contents)
    if m := MODIFIED_DATE_PATTERN.fullmatch(line):
        return (utils.to_datetime(m.group(1)), rest)
    else:
        raise ValueError("No modified date at start of string")


def get_header_version(contents: List[str]) -> Tuple[sections.Version, List[str]]:
    """Parse the header version string."""
    line, rest = _next_line(contents)
    if m := HEADER_VERSION_PATTERN.fullmatch(line):
        return (sections.Version(**m.groupdict()), rest)
    else:
        raise ValueError("No header version in string")


def get_section(
    contents: List[str], section_name: str
) -> Tuple[Dict[str, Any], List[str]]:
    """Parse a named section."""
    line, rest = _next_line(contents)
    if line != "*" + section_name.upper():
        raise ValueError(
            f"{section_name.upper()} section not present, found {line} instead"
        )
    section_info: Dict[str, Any] = {}
    last_key = None
    indent_level = 0  # pragma: no mutate
    while not utils.is_section_heading(rest[0]):
        line, rest = _next_line(rest)
        current_indent_level = len(
            list(itertools.takewhile(lambda c: c in [" "], line))
        )
        if line.strip() == "" or line.startswith("!"):
            # skip comments
            if utils.is_table_mask(line):
                raise ValueError(
                    f"'{line}' is a table mask, but the table header was not found."
                )
        elif m := TABLE_START_PATTERN.fullmatch(line):
            # handle table
            table_name = m.group(1).lower()
            # table column names
            column_names_lines = []
            line, rest = _next_line(rest)
            # column names may appear across multiple lines
            while not utils.is_table_mask(line):
                if not line.lstrip().startswith("!"):
                    raise ValueError(
                        f"'{line}' was expected to contain the column names for the table {table_name}."
                    )
                column_names_lines.append(line)
                line, rest = _next_line(rest)
            mask = line

            raw_column_names: List[List[str]] = []
            for column_names_line in column_names_lines:
                # apply column mask in case names contain spaces
                names = utils.apply_column_mask(column_names_line, mask)
                raw_column_names.append([])
                for name in names:
                    raw_column_names[-1].append(name.lower().strip())
            # combine the column names from each row
            column_names = [
                " ".join(lines).strip().replace(" ", "_")
                for lines in zip(*raw_column_names)
            ]

            # values
            section_info[table_name] = []
            line, rest = _next_line(rest)
            while not END_PATTERN.fullmatch(line.lstrip()):
                section_info[table_name].append(
                    {
                        column_names[i]: v
                        for i, v in enumerate(utils.apply_column_mask(line, mask))
                    }
                )
                line, rest = _next_line(rest)
            last_key = None
        elif m := ARRAY_START_PATTERN.fullmatch(line):
            array_name = m.group(1).lower()
            section_info[array_name] = []
            line, rest = _next_line(rest)
            while not END_PATTERN.fullmatch(line):
                section_info[array_name].append(line)
                line, rest = _next_line(rest)
            last_key = None
        elif m := REMARKS_START_PATTERN.fullmatch(line):
            # handle remarks
            remarks = []
            line, rest = _next_line(rest)
            while not END_PATTERN.fullmatch(line):
                remarks.append(line)
                line, rest = _next_line(rest)
            section_info[REMARKS] = "\n".join(remarks)  # pragma: no mutate
            last_key = None
        elif _has_key_prefix(line):
            # handle single entry
            key, value = line.split(":", 1)
            last_key = key.strip().lower()
            section_info[last_key] = value.strip()
            indent_level = current_indent_level
        elif last_key is not None and current_indent_level > indent_level:
            # handle entry continuation
            section_info[last_key] += f" {line.strip()}"
        else:
            raise ValueError(f"Unexpected text: '{line}'")
    return section_info, rest


def get_file(contents: List[str]) -> Tuple[sections.FileInfo, List[str]]:
    """Parse the \\*FILE section."""
    file_dict, rest = get_section(contents, "file")
    start_time = (
        utils.to_datetime(file_dict[START_TIME])
        if START_TIME in file_dict
        else datetime.datetime.min
    )
    end_time = (
        utils.to_datetime(file_dict[END_TIME])
        if END_TIME in file_dict
        else datetime.datetime.min
    )
    time_zero = (
        utils.to_datetime(file_dict[TIME_ZERO])
        if TIME_ZERO in file_dict
        else datetime.datetime.min
    )
    time_increment = (
        utils.to_increment(file_dict[TIME_INCREMENT].split("!")[0].strip())
        if TIME_INCREMENT in file_dict
        else datetime.timedelta(seconds=0)
    )
    time_units = file_dict[TIME_UNITS] if TIME_UNITS in file_dict else ""
    number_of_records = int(file_dict[NUMBER_OF_RECORDS])
    data_description = (
        file_dict[DATA_DESCRIPTION] if DATA_DESCRIPTION in file_dict else "n/a"
    )
    file_type = file_dict[FILE_TYPE] if FILE_TYPE in file_dict else "n/a"
    number_of_channels = int(file_dict[NUMBER_OF_CHANNELS])
    channels = (
        [sections.Channel(**elem) for elem in file_dict[CHANNELS]]
        if CHANNELS in file_dict
        else []
    )
    channel_details = (
        [sections.ChannelDetail(**elem) for elem in file_dict[CHANNEL_DETAIL]]
        if CHANNEL_DETAIL in file_dict
        else []
    )
    data_type = file_dict[DATA_TYPE] if DATA_TYPE in file_dict else "n/a"
    to_remove = " \n\t'"  # pragma: no mutate
    if FORMAT in file_dict:
        format_str = file_dict[FORMAT].strip(to_remove)
        if CONTINUED in file_dict:
            format_str += file_dict[CONTINUED].strip(to_remove)
    else:
        format_info = [
            utils.format_string(
                detail.format, detail.type, detail.width, detail.decimal_places
            )
            for detail in channel_details
        ]
        format_str = "({})".format(",".join(format_info))
    pad = float(file_dict[PAD]) if PAD in file_dict else math.nan
    remarks = file_dict[REMARKS] if REMARKS in file_dict else ""
    file_info = sections.FileInfo(
        start_time=start_time,
        end_time=end_time,
        time_zero=time_zero,
        time_increment=time_increment,
        time_units=time_units,
        number_of_records=number_of_records,
        data_description=data_description,
        file_type=file_type,
        format=format_str,
        data_type=data_type,
        pad=pad,
        number_of_channels=number_of_channels,
        channels=channels,
        channel_details=channel_details,
        remarks=remarks,
        raw=file_dict,
    )
    return file_info, rest


def get_administration(
    contents: List[str],
) -> Tuple[sections.Administration, List[str]]:
    """Parse the \\*ADMINISTRATION section."""
    admin_dict, rest = get_section(contents, "administration")
    mission = admin_dict[MISSION] if MISSION in admin_dict else "n/a"
    agency = admin_dict[AGENCY] if AGENCY in admin_dict else "n/a"
    country = admin_dict[COUNTRY] if COUNTRY in admin_dict else "n/a"
    project = admin_dict[PROJECT] if PROJECT in admin_dict else "n/a"
    scientist = admin_dict[SCIENTIST] if SCIENTIST in admin_dict else "n/a"
    platform = admin_dict[PLATFORM] if PLATFORM in admin_dict else "n/a"
    remarks = admin_dict[REMARKS] if REMARKS in admin_dict else ""
    admin_info = sections.Administration(
        mission=mission,
        agency=agency,
        country=country,
        project=project,
        scientist=scientist,
        platform=platform,
        remarks=remarks,
        raw=admin_dict,
    )
    return admin_info, rest


def get_location(contents: List[str]) -> Tuple[sections.Location, List[str]]:
    """Parse the \\*LOCATION section"""
    location_dict, rest = get_section(contents, "location")
    geographic_area = (
        location_dict[GEOGRAPHIC_AREA] if GEOGRAPHIC_AREA in location_dict else "n/a"
    )
    station = location_dict[STATION] if STATION in location_dict else "n/a"
    event_number = (
        int(location_dict[EVENT_NUMBER]) if EVENT_NUMBER in location_dict else -1
    )
    latitude = utils.get_latitude(location_dict[LATITUDE])
    longitude = utils.get_longitude(location_dict[LONGITUDE])
    water_depth = (
        float(location_dict[WATER_DEPTH])
        if WATER_DEPTH in location_dict
        and location_dict[WATER_DEPTH].lower() not in ["", "unknown"]
        else -1
    )
    remarks = location_dict[REMARKS] if REMARKS in location_dict else ""
    location_info = sections.Location(
        geographic_area=geographic_area,
        station=station,
        event_number=event_number,
        latitude=latitude,
        longitude=longitude,
        water_depth=water_depth,
        remarks=remarks,
        raw=location_dict,
    )
    return location_info, rest


def get_instrument(contents: List[str]) -> Tuple[sections.Instrument, List[str]]:
    """Parse the \\*INSTRUMENT section"""
    instrument_dict, rest = get_section(contents, "instrument")
    kind = instrument_dict[TYPE] if TYPE in instrument_dict else "n/a"
    model = instrument_dict[MODEL] if MODEL in instrument_dict else "n/a"
    serial_number = (
        instrument_dict[SERIAL_NUMBER] if SERIAL_NUMBER in instrument_dict else "n/a"
    )
    depth = float(instrument_dict[DEPTH]) if DEPTH in instrument_dict else math.nan
    remarks = instrument_dict[REMARKS] if REMARKS in instrument_dict else ""
    instrument_info = sections.Instrument(
        type=kind,
        model=model,
        serial_number=serial_number,
        depth=depth,
        remarks=remarks,
        raw=instrument_dict,
    )
    return instrument_info, rest


def get_history(contents: List[str]) -> Tuple[sections.History, List[str]]:
    """Parse the \\*HISTORY section"""
    history_dict, rest = get_section(contents, "history")
    programs = (
        [sections.Program(*elem) for elem in history_dict[PROGRAMS]]
        if PROGRAMS in history_dict
        else []
    )
    remarks = history_dict[REMARKS] if REMARKS in history_dict else ""
    history_info = sections.History(
        programs=programs,
        remarks=remarks,
        raw=history_dict,
    )
    return history_info, rest


def get_calibration(contents: List[str]) -> Tuple[sections.Calibration, List[str]]:
    """Parse the \\*CALIBRATION section"""
    calibration_dict, rest = get_section(contents, "calibration")
    corrected_channels = (
        calibration_dict[CORRECTED_CHANNELS]
        if CORRECTED_CHANNELS in calibration_dict
        else []
    )
    remarks = calibration_dict[REMARKS] if REMARKS in calibration_dict else ""
    calibration_info = sections.Calibration(
        corrected_channels=corrected_channels,
        remarks=remarks,
        raw=calibration_dict,
    )
    return calibration_info, rest


def get_raw(contents: List[str]) -> Tuple[sections.Raw, List[str]]:
    """Parse the \\*RAW section"""
    raw_dict, rest = get_section(contents, "raw")
    remarks = raw_dict[REMARKS] if REMARKS in raw_dict else ""
    raw_info = sections.Raw(
        remarks=remarks,
        raw=raw_dict,
    )
    return raw_info, rest


def get_deployment(contents: List[str]) -> Tuple[sections.Deployment, List[str]]:
    """Parse the \\*DEPLOYMENT section"""
    deployment_dict, rest = get_section(contents, "deployment")
    mission = deployment_dict[MISSION] if MISSION in deployment_dict else "n/a"
    type = deployment_dict[TYPE] if TYPE in deployment_dict else "n/a"
    anchor_dropped = (
        utils.to_datetime(deployment_dict[TIME_ANCHOR_DROPPED])
        if TIME_ANCHOR_DROPPED in deployment_dict
        else datetime.datetime.min
    )
    remarks = deployment_dict[REMARKS] if REMARKS in deployment_dict else ""
    deployment_info = sections.Deployment(
        mission=mission,
        type=type,
        anchor_dropped=anchor_dropped,
        remarks=remarks,
        raw=deployment_dict,
    )
    return deployment_info, rest


def get_recovery(contents: List[str]) -> Tuple[sections.Recovery, List[str]]:
    """Parse the \\*RECOVERY section"""
    recovery_dict, rest = get_section(contents, "recovery")
    mission = recovery_dict[MISSION] if MISSION in recovery_dict else "n/a"
    anchor_released = (
        utils.to_datetime(recovery_dict[TIME_ANCHOR_RELEASED])
        if TIME_ANCHOR_RELEASED in recovery_dict
        else datetime.datetime.min
    )
    remarks = recovery_dict[REMARKS] if REMARKS in recovery_dict else ""
    recovery_info = sections.Recovery(
        mission=mission,
        anchor_released=anchor_released,
        remarks=remarks,
        raw=recovery_dict,
    )
    return recovery_info, rest


def get_comments(contents: List[str]) -> Tuple[str, List[str]]:
    """Parse the \\*COMMENTS section"""
    line, rest = _next_line(contents)
    if COMMENTS_START_PATTERN.fullmatch(line):
        lines = []
        while not utils.is_section_heading(rest[0]):
            line, rest = _next_line(rest)
            lines.append(line)
        return "\n".join(lines), rest
    else:
        raise ValueError("No COMMENTS section found")


def _has_date(contents: str) -> bool:
    return DATE_PATTERN.search(contents) is not None


def _has_time(contents: str) -> bool:
    return TIME_PATTERN.search(contents) is not None


def _process_item(contents: Any) -> Any:
    if not isinstance(contents, str):
        return contents
    has_date = _has_date(contents)
    has_time = _has_time(contents)
    if has_date and has_time:
        return utils.to_datetime(contents)
    elif has_date:
        return utils.to_date(contents)
    elif has_time:
        return utils.to_time(contents)
    else:
        return contents.strip()


def _postprocess_line(line: List[Any]) -> List[Any]:
    return [_process_item(item) for item in line]


def get_data(contents: str, format: str, records: int) -> Tuple[List[List[Any]], str]:
    """Process the data in the file"""
    lines = contents.splitlines()
    while "" in lines:
        lines.remove("")
    if len(lines) < records:
        raise ValueError(f"Insufficient data for requested number of records")
    reader = ff.FortranRecordReader(format)
    data = [_postprocess_line(reader.read(line)) for line in lines[:records]]
    rest = "\n".join(lines[records:])  # pragma: no mutate
    return data, rest



"""This module holds constants for compiled regular expressions to be used in other modules.

Because there are so many regular expressions used with all sorts of access patterns, the internal
cache of regular expressions in re.match seems to be beyond its limit.
"""
import re


# common date and time patterns
DATE_STR = r"\d{4}[/-]\d{2}[/-]\d{2}"
DATE_PATTERN = re.compile(DATE_STR)
TIME_STR = r"\d{2}:\d{2}(:\d{2}(.\d*)?)?"
TIME_PATTERN = re.compile(TIME_STR)


# used in utils.py
TIMEZONE_STR = r"[A-Za-z]{3}"

# the names inside the < and > must be the same as the arguments to utils._to_datetime()
MATCH_DATE = f"(?P<date>{DATE_STR})"
MATCH_TZ = f"(?P<tz>{TIMEZONE_STR})"
MATCH_TIME = f"(?P<time>{TIME_STR})"

DATE_TIME_PATTERN = re.compile(f"{MATCH_DATE} {MATCH_TIME}")
TZ_DATE_TIME_PATTERN = re.compile(f"{MATCH_TZ} {MATCH_DATE}( {MATCH_TIME})?")
DATE_TIME_TZ_PATTERN = re.compile(f"{MATCH_DATE}( {MATCH_TIME})? {MATCH_TZ}")

SECTION_HEADING_PATTERN = re.compile(r"\*[A-Z ]+")


# used in parsing.py
MODIFIED_DATE_PATTERN = re.compile(rf"\*({DATE_STR} {TIME_STR})")

# the names inside the < and > must be the same as the arguments to sections.Version.__init__()
MATCH_VERSION = r"(?P<version_no>\d+.\d+)"
MATCH_DATE1 = f"(?P<date1>{DATE_STR})"
MATCH_DATE2 = f"(?P<date2>{DATE_STR})"
MATCH_TAG = "(?P<tag>[a-zA-Z0-9.]+)"

HEADER_VERSION_PATTERN = re.compile(
    rf"\*IOS HEADER VERSION +{MATCH_VERSION} +{MATCH_DATE1}( +{MATCH_DATE2}( +{MATCH_TAG})?)?"
)

KEY_PREFIX_PATTERN = re.compile(r"[^:;!?]*")

TABLE_START_PATTERN = re.compile(r"\s*\$TABLE: (.+)")
ARRAY_START_PATTERN = re.compile(r"\s*\$ARRAY: (.+)")
REMARKS_START_PATTERN = re.compile(r"\s*\$REMARKS?")
END_PATTERN = re.compile(r"\s*\$END")

COMMENTS_START_PATTERN = re.compile(r"\*COMMENTS")

END_OF_HEADER_PATTERN = re.compile(r"\*END OF HEADER")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ios_shell']

package_data = \
{'': ['*']}

install_requires = \
['fortranformat>=1.0.1,<2.0.0']

extras_require = \
{'pandas': ['pandas>=1.4.2,<2.0.0']}

setup_kwargs = {
    'name': 'ios-shell',
    'version': '1.2.2',
    'description': "Parses files formatted by the Institute of Ocean Sciences' IOSShell tool",
    'long_description': '# IOS Shell file parsing library\n\nThis is a library for parsing files in the IOS Shell format.\nIt has no relation to iOS or terminal shells.\n\n## Prerequisites\n\n- Python 3.8, 3.9, or 3.10 with pip\n\n## Example\n\nIf `1930-003-0058.bot` is in the same folder as the program using ios_shell:\n\n    parsed_file = ios_shell.ShellFile.fromfile("1930-003-0058.bot")\n\nTo use the contents of a file previously read into the program:\n\n    parsed_contents = ios_shell.ShellFile.fromcontents(contents)\n',
    'author': 'James Hannah',
    'author_email': 'jhannah@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cyborgsphinx/ios-shell',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
